/*
 * puzzle.c
 *
 *
 * reconstruction of tree topologies with quartet puzzling
 *
 * by Korbinian Strimmer and Arndt von Haeseler (1995)
 * this file belongs to the Puzzle version 2.2 distribution
 *
 */


#define MAIN_MODULE 1

#include "puzzle.h"



/******************************************************************************/
/* random numbers generator  (Numerical recipes)                              */
/******************************************************************************/


double randomunitintervall()
/* Long period (> 2e18) random number generator. Returns a uniform random
   deviate between 0.0 and 1.0 (exclusive of endpoint values).

   Source: Press et al., "Numerical recipes", Cambridge University Press */
{
	int j;
	long k;
	static long idum2=123456789;
	static long iy=0;
	static long iv[NTAB];
	float temp;

	if (idum <= 0) {
		if (-(idum) < 1)
			idum=1;
		else
			idum=-(idum);
		idum2=(idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(idum)/IQ1;
			idum=IA1*(idum-k*IQ1)-k*IR1;
			if (idum < 0)
				idum += IM1;
			if (j < NTAB)
				iv[j] = idum;
		}
		iy=iv[0];
	}
	k=(idum)/IQ1;
	idum=IA1*(idum-k*IQ1)-k*IR1;
	if (idum < 0)
		idum += IM1;
	k=idum2/IQ2;
	idum2=IA2*(idum2-k*IQ2)-k*IR2;
	if (idum2 < 0)
		idum2 += IM2;
	j=iy/NDIV;
	iy=iv[j]-idum2;
	iv[j] = idum;
	if (iy < 1)
		iy += IMM1;
	if ((temp=AM*iy) > RNMX)
		return RNMX;
	else
		return (double) temp;
}


void initrandom()
{
	srand((unsigned) time(NULL));
	idum=-(long) rand();
}


/* returns a random integer in the range [0; n - 1] */
int randominteger(int n)
{
	return (int) floor(randomunitintervall()*n);
}



/******************************************************************************/
/* functions for representing and building trees                                           */
/******************************************************************************/ 


/* initialize tree with the following starting configuration

                 2
          0  +------- C(=2)
  A(=0) -----+
             +------- B(=1)
                 1
                                */
void inittree()
{
	int i, maxbrnch;
	
	
	/* number of edges */
	maxbrnch = 2 * Maxspc - 3;
	
	/* allocate the memory for the whole tree */
	
	/* allocate memory for vector with all the edges of the tree */
	edge = (ONEEDGE *) calloc(maxbrnch, sizeof(ONEEDGE) );
	if (edge == NULL) maerror("edge [inittree]");

	/* allocate memory for vector with edge numbers of leaves */ 
	edgeofleaf = (int *) calloc(Maxspc, sizeof(int) );
	if (edgeofleaf == NULL) maerror("edgeofleaf [inittree]");
	
	/* allocate memory for all the edges the edge map */
	for (i = 0; i < maxbrnch; i++) {
		edge[i].edgemap = (int *) calloc(maxbrnch, sizeof(int) );
		if (edge[i].edgemap == NULL) maerror("edgemap [inittree]");
	}
		
	/* number all edges */
	for (i = 0; i < maxbrnch; i++) edge[i].numedge = i;
	
	/* initialize tree */
	
	nextedge = 3;
	nextleaf = 3;
	
	/* edge maps */
	(edge[0].edgemap)[0] = 0; /* you are on the right edge */
	(edge[0].edgemap)[1] = 4; /* go down left for leaf 1 */
	(edge[0].edgemap)[2] = 5; /* go down right for leaf 2 */
	(edge[1].edgemap)[0] = 1; /* go up for leaf 0 */
	(edge[1].edgemap)[1] = 0; /* you are on the right edge */
	(edge[1].edgemap)[2] = 3; /* go up/down right for leaf 2 */
	(edge[2].edgemap)[0] = 1; /* go up for leaf 0 */
	(edge[2].edgemap)[1] = 2; /* go up/down left for leaf 1 */
	(edge[2].edgemap)[2] = 0; /* you are on the right edge */
	
	/* interconnection */
	edge[0].up = NULL;
	edge[0].downleft = &edge[1];
	edge[0].downright = &edge[2];
	edge[1].up = &edge[0];
	edge[1].downleft = NULL;
	edge[1].downright = NULL;
	edge[2].up = &edge[0];
	edge[2].downleft = NULL;
	edge[2].downright = NULL;
	
	/* edges of leaves leaves */
	edgeofleaf[0] = 0;
	edgeofleaf[1] = 1;
	edgeofleaf[2] = 2;
}


/* add next leaf on the specified edge */
void addnextleaf(int dockedge)
{	
	int i;
	

	if (dockedge >= nextedge) {
		fprintf(stderr, "\nERROR: Trying to add leaf %d", nextleaf);
		fprintf(stderr, " to nonexisting edge %d [addnextleaf]\n", dockedge);
		exit(1);
	}
	
	if (nextleaf >= Maxspc) {
		fprintf(stderr, "\nERROR: Trying to add leaf %d", nextleaf);
		fprintf(stderr, " to a tree with %d leaves [addnextleaf]\n", Maxspc);
		exit(1);
	}
		
	/* necessary change in edgeofleaf if dockedge == edgeofleaf[0] */
	if (edgeofleaf[0] == dockedge) edgeofleaf[0] = nextedge;
	
	/* adding nextedge to the tree */
	edge[nextedge].up = edge[dockedge].up;
	edge[nextedge].downleft = &edge[dockedge];
	edge[nextedge].downright = &edge[nextedge+1];
	edge[dockedge].up = &edge[nextedge];

	if (edge[nextedge].up != NULL) {
		if ( ((edge[nextedge].up)->downleft) == &edge[dockedge] ) 
			(edge[nextedge].up)->downleft  = &edge[nextedge];
		else   
			(edge[nextedge].up)->downright = &edge[nextedge];
	}

	/* adding nextedge + 1 to the tree */
	edge[nextedge+1].up = &edge[nextedge];
	edge[nextedge+1].downleft = NULL;
	edge[nextedge+1].downright = NULL;
	edgeofleaf[nextleaf] = nextedge+1;
	
	/* the two new edges get info about the old edges */
	/* nextedge */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* down right changes to down left */
			case 5:		(edge[nextedge].edgemap)[i] = 4;
						break;
						
			/* null changes to down left */
			case 0:		(edge[nextedge].edgemap)[i] = 4;
						break;
			
			default: 	(edge[nextedge].edgemap)[i] =
							(edge[dockedge].edgemap)[i];
						break;
		}
	}

	/* nextedge + 1 */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* up/down left changes to up */
			case 2:		(edge[nextedge+1].edgemap)[i] = 1;
						break;
			
			/* up/down right changes to up */		
			case 3:		(edge[nextedge+1].edgemap)[i] = 1;
						break;
						
			/* down left changes to up/down left */			
			case 4:		(edge[nextedge+1].edgemap)[i] = 2;
						break;

			/* down right changes to up/down left */	
			case 5:		(edge[nextedge+1].edgemap)[i] = 2;
						break;
			
			/* null changes to up/down left */
			case 0:		(edge[nextedge+1].edgemap)[i] = 2;
						break;
			
			/* up stays up */
			default: 	(edge[nextedge+1].edgemap)[i] =
							(edge[dockedge].edgemap)[i];
						break;
		}
	}

	/* dockedge */
	for (i = 0; i < nextedge; i++) {
		switch ( (edge[dockedge].edgemap)[i] ) {
			
			/* up/down right changes to up */
			case 3:		(edge[dockedge].edgemap)[i] = 1;
						break;
						
			/* up/down left changes to up */
			case 2:		(edge[dockedge].edgemap)[i] = 1;
						break;
			
			default: 	break;
		}
	}

	/* all edgemaps are updated for the two new edges */
	/* nextedge */
	(edge[nextedge].edgemap)[nextedge] = 0;
	(edge[nextedge].edgemap)[nextedge+1] = 5; /* down right */
	
	/* nextedge + 1 */
	(edge[nextedge+1].edgemap)[nextedge] = 1; /* up */
	(edge[nextedge+1].edgemap)[nextedge+1] = 0;
	
	/* all other edges */
	for (i = 0; i < nextedge; i++) {
		(edge[i].edgemap)[nextedge] = (edge[i].edgemap)[dockedge];
		(edge[i].edgemap)[nextedge+1] = (edge[i].edgemap)[dockedge];
	}
	
	/* an extra for dockedge */
	(edge[dockedge].edgemap)[nextedge] = 1; /* up */
	(edge[dockedge].edgemap)[nextedge+1] = 3; /* up/down right */


	nextleaf++;
	nextedge = nextedge + 2;
}


/* free memory (to be called after inittree) */
void freetree()
{
	int i;

	for (i = 0; i < 2 * Maxspc - 3; i++) free(edge[i].edgemap);
	free(edge);
	free(edgeofleaf);	
}


/* clear all edgeinfos */
void resetedgeinfo()
{
	int i;
	
	for (i = 0; i < nextedge; i++)
		edge[i].edgeinfo = 0;
}


/* increment all edgeinfo between leaf A and B */
void incrementedgeinfo(int A, int B, int w)
{	
	int curredge, finaledge, nextstep;

	
	if (A == B) return;
	
	finaledge = edgeofleaf[B];
	
	curredge = edgeofleaf[A];
	edge[curredge].edgeinfo = edge[curredge].edgeinfo + w;
	
	while (curredge != finaledge) {
		nextstep = (edge[curredge].edgemap)[finaledge];
		switch (nextstep) {

			/* up */
			case 1:	curredge = (edge[curredge].up)->numedge;
					break;
				
			/* up/down left */
			case 2:	curredge = ((edge[curredge].up)->downleft)->numedge;
					break;

			/* up/down right */
			case 3:	curredge = ((edge[curredge].up)->downright)->numedge;
					break;
				
			/* down left */
			case 4:	curredge = (edge[curredge].downleft)->numedge;
					break;
				
			/* down right */
			case 5:	curredge = (edge[curredge].downright)->numedge;
					break;
				
		}
		edge[curredge].edgeinfo = edge[curredge].edgeinfo + w;
	}	
}


/* prototype so that writeOTU knows about itself */
void writeOTU(FILE *outfp, int ed);


/* write OTU (recursive function) writes OTU sitting on edge ed */
void writeOTU(FILE *outfp, int ed)
{	
	int i;

	/* test whether we are on a leaf */
	if (edge[ed].downright == NULL && edge[ed].downleft == NULL) {
		for (i = 1; i < nextleaf; i++) {
			if (edgeofleaf[i] == ed) { /* i is the leaf of ed */
				column += fputid(outfp, trueID[i]);
				return;
			}
		}
	}

	/* we are NOT on a leaf */
	fprintf(outfp, "(");
	column++;
	writeOTU(outfp, edge[ed].downleft->numedge);
	fprintf(outfp, ", ");
	column++;
	column++;
	if (column > 55) {
		column = 2;
		fprintf(outfp, "\n  ");
	}
	writeOTU(outfp, edge[ed].downright->numedge);	
	fprintf(outfp, ")");
	column++;
}

 
/* write tree to treefile */
void writetree(FILE *outfp)
{	
	column = 1;
	fprintf(outfp, "(");
	column += fputid(outfp, trueID[0]) + 3;
	fprintf(outfp, ", ");
	writeOTU(outfp, edge[edgeofleaf[0]].downleft->numedge);
	column++;
	column++;
	fprintf(outfp, ", ");
	writeOTU(outfp, edge[edgeofleaf[0]].downright->numedge);	
	fprintf(outfp, ");\n");	
}


/* checks which edge has the lowest edgeinfo */
void minimumedgeinfo()
{
	int i;


	minedge = 0;
	mininfo = edge[0].edgeinfo;
	for (i = 1; i < nextedge; i++)
		if (edge[i].edgeinfo < mininfo) {
			minedge = i;
			mininfo = edge[i].edgeinfo;
		}
}


/* Look at the path from chooseA to chooseC and compare its edges to the edges found
   on the path from chooseA to chooseB. The first edge on A-C that is not also found
   on A-B is the cutting edge */
int findcuttingedge()
{
	int curredge, finaledgeB, nextstepB, finaledgeC, nextstepC;
	
	
	finaledgeB = edgeofleaf[chooseB]; /* final B edge */
	finaledgeC = edgeofleaf[chooseC]; /* final C edge */
	
	curredge = edgeofleaf[chooseA]; /* start edge */

	while (curredge != finaledgeC) {
		nextstepB = (edge[curredge].edgemap)[finaledgeB];
		nextstepC = (edge[curredge].edgemap)[finaledgeC];
		switch (nextstepC) {
		
			/* up */
			case 1:	curredge = (edge[curredge].up)->numedge;
					break;

			/* up/down left */
			case 2:	curredge = ((edge[curredge].up)->downleft)->numedge;
					break;

			/* up/down right */
			case 3:	curredge = ((edge[curredge].up)->downright)->numedge;
					break;
				
			/* down left */
			case 4:	curredge = (edge[curredge].downleft)->numedge;
					break;

			/* down right */
			case 5:	curredge = (edge[curredge].downright)->numedge;
					break;

		}
		if (nextstepB != nextstepC) return curredge;
	}
	
	fprintf(stderr, "\nERROR:  No cutting edge [findcuttingedge]\n");
	exit(1);
}


/* prototype for recursive function */
void searchpartition(int i);


/* analyse bipartition */
void searchpartition(int i)
{	
	int j;


	/* edge i belongs to cluster */
	edgelist[i] = TRUE;

	if ( edge[i].downright == NULL ||
		 edge[i].downleft == NULL) { /* if i is leaf */

		/* check out what leaf j sits on this edge i */		
		for (j = 1; j < nextleaf; j++) {
			if (edgeofleaf[j] == i) {
				leaflist[j] = TRUE; /* this leaf belongs to the cluster */
				return;
			}	
		}	
	} else { /* still on inner branch */
		searchpartition(edge[i].downleft->numedge);
		searchpartition(edge[i].downright->numedge);
	}
}


/* mark all edges in the cluster that is defined by cutting edge and
   that contains chooseA and chooseB as irrelvant for further analysis.
   declare in freeleaves only the leaf chooseA as free and
   reduce numleaves correspondingly */
void markcluster()
{	
	int i;


	/* check whether cutting edge is a internal branch or not */

	if ( edgeofleaf[0] == cuttingedge ||
		 edge[cuttingedge].downright == NULL ||
		 edge[cuttingedge].downleft == NULL) { /* external branch */

		/* game over - this is the add-on edge */
		numleaves = 2;
		for (i = 0; i < nextedge; i++) edge[i].edgeinfo = 1;
		edge[cuttingedge].edgeinfo = 0; /* minimum edge */

	} else { /* internal branch */

		/* initialise variables */
		for (i = 0; i < nextleaf; i++) leaflist[i] = FALSE;
		for (i = 0; i < nextedge; i++) edgelist[i] = FALSE;
	
		searchpartition(cuttingedge);
		
		/* check whether chooseA is TRUE or FALSE */
		if (leaflist[chooseA] == TRUE) { /* cluster ok */

			for (i = 0; i < nextleaf; i++) {
				if (leaflist[i] == TRUE) {
					if (freeleaves[i] == TRUE) {
						freeleaves[i] = FALSE;
						numleaves--;
					}
				}
			}
			for (i = 0; i < nextedge; i++) {
				if (edgelist[i] == TRUE) edge[i].edgeinfo = 1;
			}
			
		} else { /* cluster must be inverted */
			for (i = 0; i < nextleaf; i++) {
				if (leaflist[i] == FALSE) {
					if (freeleaves[i] == TRUE) {
						freeleaves[i] = FALSE;
						numleaves--;
					}
				}
			}
			for (i = 0; i < nextedge; i++) {
				if (edgelist[i] == FALSE) edge[i].edgeinfo = 1;
			}
			edge[cuttingedge].edgeinfo = 0;
		}
		
		/* chooseA remains active */
		freeleaves[chooseA] = TRUE;
		numleaves++;
	}
}



/******************************************************************************/
/* storing and evaluating quartet branching information                       */
/******************************************************************************/


/* allocate memory for quartets */
void mallocquartets(int taxa)
{
	size_t numquartets, nc, numch;


	/* number of quartets */
	numquartets = (size_t) (taxa*(taxa-1)*(taxa-2)*(taxa-3)/24.);
	numch = (size_t) (ceil((double)numquartets/2.));

	/* allocate memory */
	quartetinfo = (unsigned char *) malloc(numch * sizeof(unsigned char) );
	if (quartetinfo == NULL) maerror("quartet");
	for (nc = 0; nc < numch; nc++) quartetinfo[nc] = 0;
}


/* free quartet memory */
void freequartets()
{	
	int a, b, c;


	free(quartetinfo);
}


/* read quartet info - a < b < c < d */
unsigned char readquartet(int a, int b, int c, int d)
{
	size_t qnum;


	qnum = (size_t) ((a + b*(b-1)/2.
						+ c*(c-1)*(c-2)/6.
						+ d*(d-1)*(d-2)*(d-3)/24.));
	if (qnum % 2 == 0) { /* even number */
		/* bits 0 to 3 */
		return (quartetinfo[qnum/2] & (unsigned char) 15);
	} else { /* odd number */
		/* bits 4 to 7 */
		return ((quartetinfo[(qnum-1)/2] & (unsigned char) 240)>>4);
	}
}


/* write quartet info - a < b < c < d, 0 <= info <= 15 */
void writequartet(int a, int b, int c, int d, unsigned char info)
{
	size_t qnum;


	qnum = (size_t) ((a + b*(b-1)/2.
						+ c*(c-1)*(c-2)/6.
						+ d*(d-1)*(d-2)*(d-3)/24.));
	if (qnum % 2 == 0) { /* even number */
		/* bits 0 to 3 */
		quartetinfo[qnum/2] =
			((quartetinfo[qnum/2] & (unsigned char) 240) |
			(info & (unsigned char) 15));
	} else { /* odd number */
		/* bits 4 to 7 */
		quartetinfo[(qnum-1)/2] =
			((quartetinfo[(qnum-1)/2] & (unsigned char) 15) |
			((info & (unsigned char) 15)<<4));
	}
}


/* checks out *all* possible quartets */
void computeallquartets()
{	
	int a, b, c, i, badquartet;
	double diff, maxdiff, diff12, diff13, diff23;
	int firstquartet;
	unsigned long nq, badqs;


	nq = 0;
	badqs = 0;
	firstquartet = TRUE;
	for (i = 3; i < Maxspc; i++) 
			for (a = 0; a < i - 2; a++)
					for (b = a + 1; b < i - 1; b++)
						for (c = b + 1; c < i; c++) {
							nq++;

							if(Info_optn) {
								fprintf(stderr, "\n  Checking quartet # %lu", nq);
							}
							
							if (Approx_optn) {
								d1 = quartet_alklhd(a,b,c,i);
								d2 = quartet_alklhd(a,c,b,i);
								d3 = quartet_alklhd(a,i,b,c);
							} else {
								d1 = quartet_lklhd(a,b,c,i);
								d2 = quartet_lklhd(a,c,b,i);
								d3 = quartet_lklhd(a,i,b,c);
							}

							if (d1 < d2) {
								if (d2 < d3) { /* d3 max */
									writequartet(a,b,c,i, 12);
									lowestml = 3;
									diff = d3 - d2;
								} else { /* d2 max */
									writequartet(a,b,c,i, 10);
									lowestml = 2;
									diff = d2 - (d1 > d3 ? d1 : d3);
								}
							} else { 
								if (d1 < d3) { /* d3 max */
									writequartet(a,b,c,i, 12);
									lowestml = 3;
									diff = d3 - d1;
								} else { /* d1 max */
									writequartet(a,b,c,i, 9);
									lowestml = 1;
									diff = d1 - (d2 > d3 ? d2 : d3);
								}
							}


							/* check whether this quartet a good quartet */

							diff12 = (d1-d2)*(d1-d2);
							diff13 = (d1-d3)*(d1-d3);
							diff23 = (d2-d3)*(d2-d3);
							badquartet = FALSE;

							if (diff12 < diff13) {
								if (diff12 < diff23) { /* diff12 min */
									/* lowestml should be 3 */
									if (lowestml != 3) badquartet = TRUE;
								} else { /* diff23 min */
									/* lowestml should be 1 */
									if (lowestml != 1) badquartet = TRUE;
								}
							} else {
								if (diff13 < diff23) { /* diff13 min*/
									/* lowestml should be 2 */
									if (lowestml != 2) badquartet = TRUE;
								} else { /* diff23 min */
									/* lowestml should be 1 */
									if (lowestml != 1) badquartet = TRUE;
								}
							}
		
							/* clear highest bit if bad quartet */
							if (badquartet) {
								writequartet(a,b,c,i, (readquartet(a,b,c,i) & 7));
								badqs++;
								if(Info_optn) {
									fprintf(stderr, "  [bad quartet: ");
									fputid(stderr, a);
									fprintf(stderr, "-");
									fputid(stderr, b);
									fprintf(stderr, "-");
									fputid(stderr, c);
									fprintf(stderr, "-");
									fputid(stderr, i);
									fprintf(stderr, "]");
								}
							};
							
							if (firstquartet) {
								firstquartet = FALSE;
								maxdiff = diff;
								startA = a;
								startB = b;
								startC = c;
								startD = i;
							}
							
							if (!badquartet) {
								if (diff > maxdiff) {
									maxdiff = diff;
									startA = a;
									startB = b;
									startC = c;
									startD = i;	
								}
							}		
						}
	if(Info_optn) fprintf(stderr, "\n");
	fprintf(stderr, "\nINFO: %lu quartets analysed - %lu bad quartets found (= %.1f %%)\n",
		nq, badqs, (double) badqs / (double) nq * 100);
}


/* check which two *leaves* out of A, B, C are closer related to each
   other than to leaf I. These two are chooseA and chooseB 
   The relationship between leaves and taxa is defined by trueID
   If the respective quartet has not yet evaluated this will done here
   lowestml is a number 1, 2, 3 if d1, d2, d3 has the lowest ML
   checkquartet returns whether a quartet is "good" or "bad"
   (checkquartet returns TRUE if the quartet is "bad") */
int checkquartet(int A, int B, int C, int I)
{	
	int input[4][2], nochange, z, ipos;
	unsigned char qresult;
	int a, b, c, i, badquartet;
	double diff12, diff13, diff23;
	
	
	input[0][0] = trueID[A];  /* taxon number */
	input[0][1] = A;          /* leaf number */
	input[1][0] = trueID[B];
	input[1][1] = B;
	input[2][0] = trueID[C];
	input[2][1] = C;
	input[3][0] = trueID[I];
	input[3][1] = I;


	/* sort after trueIDs */
	do {
		nochange = TRUE;
		if (input[2][0] > input[3][0]) {
			z = input[3][0];
			input[3][0] = input[2][0];
			input[2][0] = z;
			z = input[3][1];
			input[3][1] = input[2][1];
			input[2][1] = z;
			nochange = FALSE;
		}
		if (input[1][0] > input[2][0]) {
			z = input[2][0];
			input[2][0] = input[1][0];
			input[1][0] = z;
			z = input[2][1];
			input[2][1] = input[1][1];
			input[1][1] = z;
			nochange = FALSE;
		}
		if (input[0][0] > input[1][0]) {
			z = input[1][0];
			input[1][0] = input[0][0];
			input[0][0] = z;
			z = input[1][1];
			input[1][1] = input[0][1];
			input[0][1] = z;
			nochange = FALSE;
		}
	} while (nochange == FALSE);

	
	/* where is leaf I ? */
	ipos = 0;
	while (input[ipos][1] != I) ipos++;


	/* look at taxon quartet */
	qresult = readquartet(input[0][0], input[1][0], input[2][0], input[3][0]);
	
	/* check whether quartet branching pattern has yet to be computed  */
	if (qresult == 0) {
		
		a = input[0][0]; /* sorted taxon numbers */
		b = input[1][0];
		c = input[2][0];
		i = input[3][0];
		
		/* compute ml for all three topologies */
		if (Approx_optn) {
				d1 = quartet_alklhd(a,b,c,i);
				d2 = quartet_alklhd(a,c,b,i);
				d3 = quartet_alklhd(a,i,b,c);
		} else {
				d1 = quartet_lklhd(a,b,c,i);
				d2 = quartet_lklhd(a,c,b,i);
				d3 = quartet_lklhd(a,i,b,c);
		}

		/* choose topology */
		if (d1 < d2) {
			if (d2 < d3) { /* d3 max */
				writequartet(a,b,c,i,12);
				qresult = 12;
				lowestml = 3;
			} else { /* d2 max */
				writequartet(a,b,c,i,10);
				qresult = 10;
				lowestml = 2;
			}
		} else { 
			if (d1 < d3) { /* d3 max */
				writequartet(a,b,c,i,12);
				qresult = 12;
				lowestml = 3;
			} else { /* d1 max */
				writequartet(a,b,c,i,9);
				qresult = 9;
				lowestml = 1;
			}
		}

		/* check whether this quartet a good quartet */

		diff12 = (d1-d2)*(d1-d2);
		diff13 = (d1-d3)*(d1-d3);
		diff23 = (d2-d3)*(d2-d3);
		badquartet = FALSE;

		if (diff12 < diff13) {
			if (diff12 < diff23) { /* diff12 min */
				/* lowestml should be 3 */
				if (lowestml != 3) badquartet = TRUE;
			} else { /* diff23 min */
				/* lowestml should be 1 */
				if (lowestml != 1) badquartet = TRUE;
			}
		} else {
			if (diff13 < diff23) { /* diff13 min*/
				/* lowestml should be 2 */
				if (lowestml != 2) badquartet = TRUE;
			} else { /* diff23 min */
				/* lowestml should be 1 */
				if (lowestml != 1) badquartet = TRUE;
			}
		}

		/* clear highest bit if bad quartet */
		if (badquartet) {
			qresult = (qresult & 7);
			writequartet(a,b,c,i, qresult);
		};
	}

	/* is this quartet good or bad ? */
	if ((qresult & 8) == 0) { /* bad quartet */
		badquartet = TRUE;
		/* highest bit is set just for the next decision */
		qresult = (qresult | 8);
	} else { /* good quartet */
		badquartet = FALSE;
	}

	/* chooseA and chooseB */
	/* chooseC is the leaf that joins I */
	                 
	switch (qresult) {

		/* 1100 */
		case 12:	if (ipos == 0 || ipos == 3) {
						chooseA = input[1][1];
						chooseB = input[2][1];
						if (ipos == 0) chooseC = input[3][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[3][1];
						if (ipos == 1) chooseC = input[2][1];
						else chooseC = input[1][1];
					}
					break;

		/* 1010 */
		case 10:	if (ipos == 0 || ipos == 2) {
						chooseA = input[1][1];
						chooseB = input[3][1];
						if (ipos == 0) chooseC = input[2][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[2][1];
						if (ipos == 1) chooseC = input[3][1];
						else chooseC = input[1][1];
					}
					break;

		/* 1001 */
		case 9:		if (ipos == 0 || ipos == 1) {
						chooseA = input[2][1];
						chooseB = input[3][1];
						if (ipos == 0) chooseC = input[1][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[1][1];
						if (ipos == 2) chooseC = input[3][1];
						else chooseC = input[2][1];
					}
					break;
		
		default:	fprintf(stderr, "\nERROR: Program error [checkquartet]\n");
					exit(1);
	}
	
	return badquartet;
}


/* check which two *leaves* out of A, B, C are closer related to each
   other than to leaf I. These two are chooseA and chooseB 
   The relationship between leaves and taxa is defined by trueID
   !!!!!!! The quartet will *ALWAYS* be recomputed (no storage) !!!!!!
   lowestml is a number 1, 2, 3 if d1, d2, d3 has the lowest ML
   checkquartet2 returns whether a quartet is "good" or "bad"
   (checkquartet2 returns TRUE if the quartet is "bad") */
int checkquartet2(int A, int B, int C, int I)
{	
	int input[4][2], nochange, z, ipos;
	unsigned char qresult;
	int a, b, c, i, badquartet;
	double diff12, diff13, diff23;


	input[0][0] = trueID[A];  /* taxon number */
	input[0][1] = A;          /* leaf number */
	input[1][0] = trueID[B];
	input[1][1] = B;
	input[2][0] = trueID[C];
	input[2][1] = C;
	input[3][0] = trueID[I];
	input[3][1] = I;


	/* sort after trueIDs */
	do {
		nochange = TRUE;
		if (input[2][0] > input[3][0]) {
			z = input[3][0];
			input[3][0] = input[2][0];
			input[2][0] = z;
			z = input[3][1];
			input[3][1] = input[2][1];
			input[2][1] = z;
			nochange = FALSE;
		}
		if (input[1][0] > input[2][0]) {
			z = input[2][0];
			input[2][0] = input[1][0];
			input[1][0] = z;
			z = input[2][1];
			input[2][1] = input[1][1];
			input[1][1] = z;
			nochange = FALSE;
		}
		if (input[0][0] > input[1][0]) {
			z = input[1][0];
			input[1][0] = input[0][0];
			input[0][0] = z;
			z = input[1][1];
			input[1][1] = input[0][1];
			input[0][1] = z;
			nochange = FALSE;
		}
	} while (nochange == FALSE);


	/* where is leaf I ? */
	ipos = 0;
	while (input[ipos][1] != I) ipos++;


	/* compute quartet branching pattern  */
	a = input[0][0]; /* sorted taxon numbers */
	b = input[1][0];
	c = input[2][0];
	i = input[3][0];

	/* compute ml for all three topologies */
	if (Approx_optn) {
			d1 = quartet_alklhd(a,b,c,i);
			d2 = quartet_alklhd(a,c,b,i);
			d3 = quartet_alklhd(a,i,b,c);
	} else {
			d1 = quartet_lklhd(a,b,c,i);
			d2 = quartet_lklhd(a,c,b,i);
			d3 = quartet_lklhd(a,i,b,c);
	}

	/* choose topology */
	if (d1 < d2) {
		if (d2 < d3) { /* d3 max */
			qresult = 12;
			lowestml = 3;
		} else { /* d2 max */
			qresult = 10;
			lowestml = 2;
		}
	} else { 
		if (d1 < d3) { /* d3 max */
			qresult = 12;
			lowestml = 3;
		} else { /* d1 max */
			qresult = 9;
			lowestml = 1;
		}
	}

	/* check whether this quartet a good quartet */

	diff12 = (d1-d2)*(d1-d2);
	diff13 = (d1-d3)*(d1-d3);
	diff23 = (d2-d3)*(d2-d3);
	badquartet = FALSE;

	if (diff12 < diff13) {
		if (diff12 < diff23) { /* diff12 min */
			/* lowestml should be 3 */
			if (lowestml != 3) badquartet = TRUE;
		} else { /* diff23 min */
			/* lowestml should be 1 */
			if (lowestml != 1) badquartet = TRUE;
		}
	} else {
		if (diff13 < diff23) { /* diff13 min*/
			/* lowestml should be 2 */
			if (lowestml != 2) badquartet = TRUE;
		} else { /* diff23 min */
			/* lowestml should be 1 */
			if (lowestml != 1) badquartet = TRUE;
		}
	}
	

	/* chooseA and chooseB */
	/* chooseC is the leaf that joins I */

	switch (qresult) {
		
		/* 1100 */
		case 12:	if (ipos == 0 || ipos == 3) {
						chooseA = input[1][1];
						chooseB = input[2][1];
						if (ipos == 0) chooseC = input[3][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[3][1];
						if (ipos == 1) chooseC = input[2][1];
						else chooseC = input[1][1];
					}
					break;
		
		/* 1010 */
		case 10:	if (ipos == 0 || ipos == 2) {
						chooseA = input[1][1];
						chooseB = input[3][1];
						if (ipos == 0) chooseC = input[2][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[2][1];
						if (ipos == 1) chooseC = input[3][1];
						else chooseC = input[1][1];
					}
					break;

		/* 1001 */
		case 9:		if (ipos == 0 || ipos == 1) {
						chooseA = input[2][1];
						chooseB = input[3][1];
						if (ipos == 0) chooseC = input[1][1];
						else chooseC = input[0][1];
					} else {
						chooseA = input[0][1];
						chooseB = input[1][1];
						if (ipos == 2) chooseC = input[3][1];
						else chooseC = input[2][1];
					}
					break;
		
		default:	fprintf(stderr, "\nERROR: Program error [checkquartet2]\n");
					exit(1);
	}
	
	return badquartet;
}



/******************************************************************************/
/* main part                                                                  */
/******************************************************************************/


/* set options interactively */
void setoptions()
{	
	char inputline[10];	


	/* defaults */
	Info_optn   = FALSE; /* show status of computation       */
	pmode_optn  = 0;     /* original quartet puzzling mode   */
	save_optn   = TRUE;  /* save quartets (fast mode)        */
	Approx_optn = TRUE;  /* approximate likelihoods          */
	Numtrial = 100;      /* number of quartet puzzling trees */

#ifdef NUC	
	AlphaYR = 1.0;
	AlphaBeta = 4.0;
	Frequ_optn = TRUE;
#endif

#ifdef PROT
	Dayhf_optn = FALSE;
	Jtt_optn = TRUE;
	Poisn_optn = FALSE;
	Frequ_optn = FALSE;
#endif

	fprintf(stderr, "\nReconstructing tree topologies with quartet puzzling\n");

 	for (;;){
			
		fprintf(stderr, "\n\nSettings for this run of Puzzle version %s (%s):\n\n",
			VERSION, MODEL);	
		fprintf(stderr, "p           Quartet puzzling algorithm?  ");
		if (pmode_optn == 0) fprintf(stderr, "Use all quartets\n");
		if (pmode_optn == 1) fprintf(stderr, "Use all good quartets\n");
		if (pmode_optn == 2) fprintf(stderr, "Use selected quartets\n");
#ifdef NUC			
		fprintf(stderr, "m          Model of sequence evolution?  ");
		if(AlphaBeta == 1.0 && AlphaYR == 1.0) fprintf(stderr, "%s\n",
			(Frequ_optn ? "F (Felsenstein 1981)" : "JC (Jukes-Cantor 1969)"));
		if((AlphaBeta != 1.0 && AlphaYR == 1.0)) fprintf(stderr, "%s\n",
			(Frequ_optn ? "HKY (Hasegawa et al. 1985)" : "K2P (Kimura 1980)"));	
		if(AlphaYR != 1.0) fprintf(stderr, "%s\n", "TN (Tamura-Nei 1993)");
		if (AlphaYR != 1.0 || AlphaBeta != 1.0)
			fprintf(stderr, "t        Transition/transversion ratio?  %.2f\n",
				AlphaBeta / 2);	
		if (AlphaYR != 1.0)
			fprintf(stderr, "r      Y-transition/R-transition ratio?  %.2f\n",
				AlphaYR);
#endif

#ifdef PROT
		fprintf(stderr, "m          Model of sequence evolution?  ");
		if (Dayhf_optn) fprintf(stderr, "Dayhoff (Dayhoff et al. 1978)\n");	
		if (Jtt_optn) fprintf(stderr, "JTT (Jones et al. 1992)\n");
		if (Poisn_optn) fprintf(stderr, "%s\n",
			(Frequ_optn ? "F (Felsenstein 1981)" : "JC (Jukes-Cantor 1969)"));
#endif

		fprintf(stderr, "f       Use empirical base frequencies?  %s\n",
			(Frequ_optn ? "Yes" : "No") );	
		fprintf(stderr, "a       Approximate maximum likelihood?  %s\n",
				(Approx_optn ? "Yes" : "No"));
		fprintf(stderr, "n     Number of quartet puzzling trees?  %d\n",
				Numtrial);
		if (pmode_optn == 2) {
		fprintf(stderr, "s             Store quartet topologies?  %s\n",
				(save_optn ? "Yes" : "No"));
		}
		fprintf(stderr, "i          Show details of computation?  %s\n",
				(Info_optn ? "Yes" : "No"));
		
		fprintf(stderr, "\nStart with analysis? ");
		fprintf(stderr, "(type y or choose from the menu) ");
		
		gets(inputline);

		if(tolower(inputline[0]) == 'y')
		break;

		switch (tolower(inputline[0])) {

			case 'a':	Approx_optn = 1 - Approx_optn;
						break;
			
			case 'f':	Frequ_optn = 1 - Frequ_optn;
						break;
			
			case 'i':	Info_optn = 1 - Info_optn;
						break;

			case 'n':	fprintf(stderr, "\n\nNumber of quartet puzzling trees: ");
						scanf("%d", &Numtrial);
						gets(inputline);
						if (Numtrial < 1) {
							fprintf(stderr,
								"\n\nAt least one tree has to be generated!\n");
							Numtrial = 100;
						}
						break;

			case 'p':	pmode_optn++;
						if (pmode_optn > 2) pmode_optn = 0;
						break;

			case 's':	if (pmode_optn == 2) {
							save_optn = 1 - save_optn;
							break;
						} else {
							fprintf(stderr,"\n\nThis is not a possible option!\n");
							break;
						}
						break;

#ifdef NUC			
			case 'm':	if(AlphaBeta == 1.0 && AlphaYR == 1.0) {/* -> K2P/HKY */
							AlphaBeta = 4.0;
							AlphaYR = 1.0;
							break;
						}
						if(AlphaBeta != 1.0 && AlphaYR == 1.0) {/* -> TN */
							AlphaBeta = 4.0;
							AlphaYR = 2.0;
							break;
						}
						if(AlphaYR != 1.0) { /* -> JC */
							AlphaBeta = 1.0;
							AlphaYR = 1.0;
							break;
						}
						break;
						
			case 't':	if (AlphaYR == 1.0 && AlphaBeta == 1.0) {
							fprintf(stderr,"\n\nThis is not a possible option!\n");
							break;
						} else {
							fprintf(stderr, "\n\nExpected transition/transversion ratio: ");
							scanf("%lf", &AlphaBeta);
							AlphaBeta = 2 * AlphaBeta;
							gets(inputline);
							if (AlphaBeta < 0) {
								fprintf(stderr, "\n\nNo negative values for the transition/transversion ratio allowed!\n");
								AlphaBeta = 4.0;
							}
							if (AlphaBeta == 0) {
								fprintf(stderr, "\n\nNo zero value for the transition/transversion ratio allowed!\n");
								AlphaBeta = 4.0;
							}
							break;
						}
						break;

			case 'r':	if (AlphaYR == 1.0) {
							fprintf(stderr,"\n\nThis is not a possible option!\n");
							break;
						} else {
							fprintf(stderr, "\n\nPyrimidine transition/Purin transition ratio: ");
							scanf("%lf", &AlphaYR);
							gets(inputline);
							if (AlphaYR < 0) {
								fprintf(stderr, "\n\nNo negative values for the Y-transition/R-transition ratio allowed!\n");
								AlphaYR = 2.0;
							}
							if (AlphaYR == 0) {
								fprintf(stderr, "\n\nNo zero value for the Y-transition/R-transition ratio allowed!\n");
								AlphaYR = 2.0;
							}
							break;
						}
						break;
#endif

#ifdef PROT
			case 'm':	if (Dayhf_optn) { /* -> JC/F */
							Dayhf_optn = FALSE;
							Jtt_optn = FALSE;
							Poisn_optn = TRUE;
							break;
						}
						if (Poisn_optn) { /* -> JTT model */
							Dayhf_optn = FALSE;
							Jtt_optn = TRUE;
							Poisn_optn = FALSE;
							break;
						}
						if (Jtt_optn) { /* -> Dayhoff model */
							Dayhf_optn = TRUE;
							Jtt_optn = FALSE;
							Poisn_optn = FALSE;
							break;
						}
						break;

#endif

		default:	fprintf(stderr, "\n\nThis is not a possible option!\n");
		}
	}
	fprintf(stderr, "\n\n");
}


/* open all needed files */
void opentreefile()
{	
	char inputline[64];

	fprintf(stderr, "\n\nPlease enter the name for the output file: ");
	gets(inputline);
	if ((treefp = fopen(inputline, "w")) == NULL) {
		fprintf(stderr, "\nERROR: I can't create the output file: %s\n", inputline);
		fprintf(stderr, "Please enter another name for this file: ");
		gets(inputline);
		if ((treefp = fopen(inputline, "w")) == NULL) {
			fprintf(stderr, "\nERROR: I can't create this file either!\n");
			exit(1);
		}
	}
	fprintf(stderr, "\n");
}


/* close needed files */
void closetreefile()
{	
	fclose(treefp);
}


main()
{	
	int i, j, a, b, c, k, ID, ii, jj, badquartet, badtree, numbadtrees;
	unsigned long nqused, nqanalysed, trial;
	time_t starttime, stoptime;
	double diff12, diff13, diff23;
#ifdef NUC
	double AlphaYBeta, AlphaRBeta, PiA, PiC, PiG, PiT, PiR, PiY;
#endif


	/* set options in interaction with the user */
	setoptions();

	/* initialize random numbers generator */
	initrandom();
	
	/* initialise ML */
	fprintf(stderr, "\nINFO: Reading and processing sequence input data\n");
	qmlstart();

#ifdef NUC	
	/* compute transition-transversion ratio */
	PiA = Freqtpm[2];
	PiC = Freqtpm[1];
	PiG = Freqtpm[3];
	PiT = Freqtpm[0];
	PiR = PiA + PiG;
	PiY = PiC + PiT;
	AlphaRBeta = 2 * AlphaBeta / (1 + AlphaYR);
	AlphaYBeta = AlphaRBeta * AlphaYR;	
	fprintf(stderr, "\nINFO: Current expected transition/transversion ratio is %.2f\n",
		(AlphaRBeta * PiA * PiG + AlphaYBeta * PiC * PiT) / (PiR * PiY) );
#endif

	/* open treefile */
	opentreefile();

	/* variables used for randomizing input order */
	trueID = new_ivector(Maxspc);
	freeID = new_ivector(Maxspc);
	
	/* variable used by fast Puzzle algorithm */
	if (pmode_optn == 2) {
		freeleaves = new_ivector(Maxspc);
		leaflist = new_ivector(Maxspc);
		edgelist = new_ivector(2*Maxspc - 3);
	}

	/* allocate memory for quartets */
	if ( !(pmode_optn == 2) || save_optn) {
		fprintf(stderr, "\nINFO: Preparing for quartet information storage\n");
		mallocquartets(Maxspc);
	}

	/* start timer */
	time(&starttime);
		
	/* compute quartets */
	if (!(pmode_optn == 2)) {
		fprintf(stderr, "\nINFO: Computing quartet maximum likelihood values\n");
		computeallquartets();
	}
	
	fprintf(stderr, "\nINFO: Starting quartet puzzling analysis\n");
	
	if (pmode_optn == 2) numbadtrees = 0; /* counter of bad trees */
	
	for (Currtrial = 0; Currtrial < Numtrial; Currtrial++) {
	
		/* assume that the next tree is good */
		if (pmode_optn == 2) badtree = FALSE;
				
		/* randomize input order */
				
		for (k = 0; k < Maxspc; k++) freeID[k] = TRUE;
		if (pmode_optn == 1) { /* special care for first quartet */
			trueID[0] = startA;
			freeID[startA] = FALSE;
			trueID[1] = startB;
			freeID[startB] = FALSE;
			trueID[2] = startC;
			freeID[startC] = FALSE;
			trueID[3] = startD;
			freeID[startD] = FALSE;
			for (k = 4; k < Maxspc; k++) {
				/* random integer in the range [0, Maxspc-1-k] */
				ID = randominteger(Maxspc - k);
				i = -1;
				j = -1;
				do {
					i++;
					if (freeID[i] == TRUE) j++; 
				} while ( j != ID);
				trueID[k] = i;
				freeID[i] = FALSE;
			}
		} else {
			for (k = 0; k < Maxspc; k++) {
				/* random integer in the range [0, Maxspc-1-k] */
				ID = randominteger(Maxspc - k);
				i = -1;
				j = -1;
				do {
					i++;
					if (freeID[i] == TRUE) j++; 
				} while ( j != ID);
					trueID[k] = i;
					freeID[i] = FALSE;
			}
		}


		/* initialize tree */
		inittree();
		if (Info_optn) {
			fprintf(stderr, "\n  Quartet puzzling tree # %d:\n",
			Currtrial + 1);
			fprintf(stderr, "   leaf 0  > ");
			fputid10(stderr, trueID[0]);
			fprintf(stderr, " <  added  [0/0 quartets used/analysed]\n");
			fprintf(stderr, "   leaf 1  > ");
			fputid10(stderr, trueID[1]);
			fprintf(stderr, " <  added  [0/0 quartets used/analysed]\n");
			fprintf(stderr, "   leaf 2  > ");
			fputid10(stderr, trueID[2]);
			fprintf(stderr, " <  added  [0/0 quartets used/analysed]\n");
		}

		/* adding all other leafs */
		for (i = 3; i < Maxspc; i++) {

			/* determine place for next leaf */ 
					
			/* clear all edgeinfos */
			resetedgeinfo();
				
			/* clear counter of quartets */
			nqanalysed = 0;
			nqused = 0;


			/*
			 * core of quartet puzzling algorithm
			 */

			if (pmode_optn == 0) { /* use all quartets (original Puzzle algorithm) */

				for (a = 0; a < nextleaf - 2; a++)
					for (b = a + 1; b < nextleaf - 1; b++)
						for (c = b + 1; c < nextleaf; c++) {

							/*	check which two leaves out of a, b, c
								are closer related to each other than
								to leaf i. These two leaves are chooseA and chooseB */

								checkquartet(a, b, c, i);
								nqanalysed++;

								incrementedgeinfo(chooseA, chooseB, 1);
								nqused++;
						}

			}
				
			if (pmode_optn == 1) { /* use all good quartets + special care for 1st quartet */

				for (a = 0; a < nextleaf - 2; a++)
					for (b = a + 1; b < nextleaf - 1; b++)
						for (c = b + 1; c < nextleaf; c++) {
						
							/*	check which two leaves out of a, b, c
								are closer related to each other than
								to leaf i. These two leaves are chooseA and chooseB */

								badquartet = checkquartet(a, b, c, i);
								nqanalysed++;

								if (!badquartet) {
									incrementedgeinfo(chooseA, chooseB, 1);
									nqused++;
								}
						}

			}
				
			if (pmode_optn == 2) { /* use selected quartets */

				/* leaves that are initally free to use */
				numleaves = nextleaf;
				for (ii = 0; ii < nextleaf; ii++) freeleaves[ii] = TRUE;
				for (ii = nextleaf; ii < Maxspc; ii++) freeleaves[ii] = FALSE;

				do {
					trial = 0;
					do {
						/* choose randomly from the available leaves to make a quartet xxxi */					
						a = randominteger(numleaves);
						ii = -1;
						jj = -1;
						do {
							ii++;
							if (freeleaves[ii] == TRUE) jj++; 
						} while ( jj != a);
						a = ii;
						freeleaves[a] = FALSE;
						numleaves--;
						b = randominteger(numleaves);
						ii = -1;
						jj = -1;
						do {
							ii++;
							if (freeleaves[ii] == TRUE) jj++; 
						} while ( jj != b);
						b = ii;
						freeleaves[b] = FALSE;
						numleaves--;
						c = randominteger(numleaves);
						ii = -1;
						jj = -1;
						do {
							ii++;
							if (freeleaves[ii] == TRUE) jj++; 
						} while ( jj != c);
						c = ii;
						numleaves = numleaves + 2;
						freeleaves[a] = TRUE;
						freeleaves[b] = TRUE;

						/* check this quartet abci */
						if (save_optn) /* store quartets */
							badquartet = checkquartet(a, b, c, i);
						else /* don't store quartets */
							badquartet = checkquartet2(a, b, c, i);
						nqanalysed++;
						trial++;

					} while (badquartet && trial < 20 && numleaves > 3);
					nqused++;

					if (badquartet && Info_optn) {
						fprintf(stderr, "\n     WARNING: Bad quartet [");
						fputid(stderr, trueID[a]);
						fprintf(stderr, ", ");
						fputid(stderr, trueID[b]);
						fprintf(stderr, ", ");
						fputid(stderr, trueID[c]);
						fprintf(stderr, ", ");
						fputid(stderr, trueID[i]);
						fprintf(stderr, "] used in adding leaf %d\n", i);
						if (numleaves == 3) {
							fprintf(stderr, "              because algorithm had no ");
							fprintf(stderr, "other choice (Available leaves = 3)\n");
						}
						if (trial == 20) {
							fprintf(stderr, "              because number of trials ");
							fprintf(stderr, "is exceeded (Number of trials = 20)\n");
						}
						fprintf(stderr, "\n");
					}
					if (badquartet) badtree = TRUE;

					/* look for that edge that cuts the tree into two parts:
					   this edge is the first edge on the path A-C that is different
					   from the edges on the path A-B */
					cuttingedge = findcuttingedge();

					/* mark all edges and leaves as irrelevant in the cluster that
					   contains chooseA and chooseB
					   Note: this changes edges, freeleaves, and numleaves */
					markcluster();

				} while (numleaves > 2);

			} /* "Use selected quartets" */


			/* find out which edge has the lowest edgeinfo */
			minimumedgeinfo();
				
			/* add the next leaf on minedge */
			addnextleaf(minedge);

			if(Info_optn) {
				fprintf(stderr, "   leaf %d  > ", i);
				fputid10(stderr, trueID[i]);
				fprintf(stderr, " <  added  [%lu/%lu quartets used/analysed]\n",
					nqused, nqanalysed);
			}
		}
		
		if (pmode_optn == 2) {
			if (badtree) numbadtrees++;
			if (numbadtrees < 10 && badtree) {
				if (Info_optn) fprintf(stderr, "\n");
				fprintf(stderr,
					"\nINFO: Generation of tree # %d will be repeated\n", Currtrial + 1);
				if (Info_optn) fprintf(stderr, "\n");
			}
		}
		
		/* writing treefile */
		if (pmode_optn == 2) {
			if (badtree && numbadtrees < 10) {
				Currtrial--; /* repeat tree generation up to 10 x if tree is bad */
			}
			else {
				writetree(treefp);
				numbadtrees = 0;
			}
		} else writetree(treefp);

		/* free memory before building the next tree */
		freetree();
	}

	/* close treefiles */
	closetreefile();
	fprintf(stderr, "\nINFO: All %d trees written to output file\n", Numtrial);

	/* stop timer */
	time(&stoptime);
	fprintf(stderr, "\nINFO: This run took %.0f seconds (= %.1f minutes)\n",
		(double) (stoptime - starttime), (stoptime - starttime)/60.);

	/* finish ML */
	qmlfinish();

	/* free memory */
	if ((!pmode_optn == 2) || save_optn) freequartets();
	free_ivector(trueID);
	free_ivector(freeID);
	if (pmode_optn == 2) {
		freeleaves = new_ivector(Maxspc);
		leaflist = new_ivector(Maxspc);
		edgelist = new_ivector(2*Maxspc - 3);
	}

	return;
}
